#### XINI.pm - eXtended INI file handling

package XINI;
push @::VERSIONS, '$Id: XINI.pm,v 1.1 2006/07/11 21:46:18 digg Exp $';

use strict;
use Carp;
require Exporter;
use vars qw(@ISA @EXPORT);
@ISA = qw(Exporter);
@EXPORT = qw(init_ini merge_ini ini ini_got ini_sections
             ini_either ini_combine ini_push ini_pop ini_raw
             ini_inheritance ini_age);

my $inihash = undef;
my @inifilenames;
my @inistack; # current section stack

sub parse_ini ($$@) {
    my ($shr, $filename, @lines) = @_;
    my ($csn, $khr) = undef;
    my %seen; # per-file
    for my $i (0 ... $#lines) {
        my $line = $lines[$i];
        chomp $line;
        if ($line =~ m/^\s*;/ or $line =~ m/^\s*$/) {
            # do nothing
        } elsif ($line =~ m/^\s*\[\s*(.*?)\s*\]\s*$/) {
            $csn = $1;
            croak "$filename:".($i + 1).": empty section name, aborting"
                if $csn =~ m/^\s*$/;
            croak "$filename:".($i + 1).": invalid section name, aborting"
                if $csn =~ m/[\$:=]/;
            croak "$filename:".($i + 1).": duplicate section name [$csn], aborting"
                if exists $seen{"$csn:"};
            $seen{"$csn:"} = 1;
            $shr->{$csn} = {} unless exists $shr->{$csn};
            $khr = $shr->{$csn};
        } elsif ($line =~ m/^\s*([^=]*?)\s*=\s*(.*)$/) {
            my ($key, $value) = ($1, $2);
            croak "$filename:".($i + 1).": key outside section, aborting"
                unless defined $khr;
            croak "$filename:".($i + 1).": empty key name, aborting"
                if $key =~ m/^\s*$/;
            croak "$filename:".($i + 1).": invalid key name, aborting"
                if $key =~ m/[:=]/;
            croak "$filename:".($i + 1).": duplicate key [$csn]$key, aborting"
                if exists $seen{"$csn:$key"};
            $seen{"$csn:$key"} = 1;
            $khr->{$key} = $value;
        } else {
            croak "$filename:".($i + 1).": ini parse error, aborting";
        }
    }
}

sub init_ini (@) {
    my (@filenames) = @_;

    my %ini;
    $inihash = \%ini;
    for my $filename (reverse @filenames) {
        my $fh;
        open $fh, '<', $filename or die "$filename: open for reading: $!";
        my @lines = <$fh>;
        close $fh;
        $lines[0] =~ s/^\xEF\xBB\xBF// if @lines; # drop the BOM
        s/\x0D?\x0A?$// for @lines;
        parse_ini $inihash, $filename, @lines;
    }
    @inifilenames = @filenames;
}

sub merge_ini ($) {
    my ($filename) = @_;

    unshift @inifilenames, $filename;
    init_ini @inifilenames;
}

# return the ancestry list, from eldest to to the mentioned section
sub ini_inheritance ($)
{
    my ($child) = @_;
    my @ancestry = ($child);
    while (1)
      {
        die "no [$ancestry[0]]" unless exists $inihash->{$ancestry[0]};
        return @ancestry unless exists $inihash->{$ancestry[0]}{'.'};
        unshift @ancestry, $inihash->{$ancestry[0]}{'.'};
      }
}

# get a single ini item, taking care of inheritance
sub ini_raw ($$) {
    my ($section, $key) = @_;
    my %visited;
AGAIN:
    return undef unless exists $inihash->{$section};
    return $inihash->{$section}{$key} if exists $inihash->{$section}{$key};
    return undef unless exists $inihash->{$section}{'.'};
    $visited{$section} = 1;
    $section = $inihash->{$section}{'.'};
    die "inheritance loop involving [$section] detected"
        if exists $visited{$section};
    goto AGAIN;
}

# Determine the generation count of $key in its section ancestry.
# If [$section]$key is found, the age is 1.
# If such $key is not found at all, returns undef.
sub ini_age ($$) {
    my ($section, $key) = @_;
    my %visited;
    my $age = 1;
AGAIN:
    return undef unless exists $inihash->{$section};
    return $age if exists $inihash->{$section}{$key};
    return undef unless exists $inihash->{$section}{'.'};
    $visited{$section} = 1;
    $section = $inihash->{$section}{'.'};
    die "inheritance loop involving [$section] detected"
        if exists $visited{$section};
    $age++;
    goto AGAIN;
}

# get list of keys in a section, taking care of inheritance
sub _ini_get_section_keys ($) {
    my ($section) = @_;
    my %visited;
    my %keys;

    while (exists $inihash->{$section}) {
        %keys = (%keys, %{$inihash->{$section}});
        last unless exists $inihash->{$section}{'.'};
        $visited{$section} = 1;
        $section = $inihash->{$section}{'.'};
        die "inheritance loop involving [$section] detected"
            if exists $visited{$section};
    }
    delete $keys{'.'};
    my @keys = sort keys %keys;
    return @keys;
}

sub _expand_var ($$) {
    my ($section, $var) = @_;
    my @sectcomp = split /\s+/, $section;
    for my $sec ($section, '$') {
        my $exp = ini_raw $sec, '$('.$var.')';
        return $exp if defined $exp;
        $exp = ini_raw $sec, '$'.$var;
        return $exp if defined $exp;
    }
    return undef;
}

# Get an ini entry.
sub ini ($) {
    my ($request) = @_;

    my ($section, $key, $mode, $default);
    if ($request =~ m/^([^\[\]=:]*?):([^\[\]=:]*?)(?:\s*(:?=)\s*(.*))?$/)
        {($section, $key, $mode, $default) = ($1, $2, $3, $4)}
    else {croak "Invalid INI request: $request"}
    if ($section eq '') {
        croak "indecisive INI request $request"
            unless @inistack;
        $section = $inistack[-1];
    }
    if (defined $mode and $mode eq ':=') {
        croak "incomplete INI change request $request"
            unless defined $section and defined $key and defined $default;
        croak "No section [$section]"
            unless exists $inihash->{$section};
        $inihash->{$section}{$key} = $default;
        return $default;
    }
    if ($key eq '' or $key =~ m/\*/ and $key !~ m/\(no star\)$/) {
        # no key given or key is mask
        croak "No section [$section]"
            unless exists $inihash->{$section};
        my @keys = _ini_get_section_keys $section;
        if ($key eq '') {
            @keys = sort grep !/^\$/, @keys;
        } else {
            my @keychars = split //, $key;
            @keychars = map {
                if ($_ eq '*') {'.*'} else {s/(\W)/\\$1/; $_}
            } @keychars;
            my $keyre = join '', @keychars;
            @keys = sort grep /^$keyre$/, @keys;
        }
        return @keys if wantarray;
        return join ' ', @keys;
    }
    $key =~ s/\s*\(no star\)$//;
    my $value = ini_raw $section, $key;
    unless (defined $value) {
        if (defined $default) {$value = $default}
        else {croak "No [$section]$key in (@inifilenames)"}
    }
    $value = [$value] unless ref $value eq 'ARRAY';
    my @results = map {
        my $value = $_;
        my $limit = 100; # avoid infinite recursion
        my $last_value;
        do {
            croak "Too deep recursion in [$section]$key"
                if --$limit == 0;
            $last_value = $value;
            $value =~ s{(\$([\w-]+\b))}{my $x = _expand_var($section, $2); defined $x ? $x : $1}ge;
            $value =~ s{(\$\(([\w-]+):([\w-]+)\))}{my $x = _expand_var($2, $3); defined $x ? $x : $1}ge;
            croak "Too long expansion for [$section]$key"
                if length $value > 65536;
        } while $value ne $last_value;
        $value;
    } @$value;
    return $results[0] if @results == 1;
    return \@results;
}

# Checks existence of a specified INI part.
# No flags, colonized interface:
# foo:bar - key 'bar' in section 'foo'
# foo:    - section 'foo'
#    :bar - key 'bar' in current section
sub ini_got ($) {
    my ($request) = @_;

    my ($section, $key, $default);
    if ($request =~ m/^([^\[\]=:]*):([^\[\]=:]*)$/)
        {($section, $key) = ($1, $2)}
    else {croak "Invalid INI request: $request"}
    croak "content-free argument to ini_got"
        if $section eq '' and $key eq '';
    if ($section eq '') {
        croak "indecisive INI request $request"
            unless @inistack;
        $section = $inistack[-1];
    }
    return exists $inihash->{$section} if $key eq '';
    return defined ini_raw $section, $key;
}

sub ini_sections (;$) {
    my ($mask) = @_;
    my @sections = keys %$inihash;
    if (defined $mask) {
        croak "invalid ini section mask '$mask'"
            unless $mask =~ m/^[^*]*\*[^*]*$/;
        $mask =~ s/([^\w*])/\\$1/g;
        $mask =~ s/\*/(.*)/;
        @sections = map {m/^$mask$/} @sections;
    }
    return @sections
}

sub ini_either ($) {
    my ($spec) = @_;
    croak "invalid ini_either spec: $spec"
        unless my ($sections, $key) = $spec =~ m/^([^\[\]:]+):([^\[\]:=]+)$/;
    my @sections = split /=/, $sections;
    my $result = undef;
    my $result_bang = undef;
    my @gots = ();
    my @gots_bang = ();
    for my $sec (@sections) {
        if (ini_got "$sec\:$key") {
            $result = ini "$sec\:$key";
            push @gots, $sec;
        }
        if (ini_got "$sec\:$key!") {
            $result_bang = ini "$sec\:$key!";
            push @gots_bang, $sec;
        }
    }
    croak "precedence conflict between (".(join '=', @gots_bang).":$key!)"
        if @gots_bang > 1;
    return $result_bang
        if @gots_bang == 1;
    croak "precedence conflict between (".(join '=', @gots).":$key)"
        if @gots > 1;
    return $result
        if @gots == 1;
    croak "neither of $spec defined"
        unless @gots;
}

sub ini_combine (@) {
    my @sections = @_;
    my %result;
    my $sections = join '=', @sections;
    for my $sec (@sections) {
        croak "invalid section name: $sec"
            unless $sec =~ m/^[^\[\]=:]+$/;
        next unless ini_got "$sec\:";
        for my $k (ini "$sec\:") {
            next if $k =~ /!$/;
            $result{$k} = ini_either "$sections\:$k";
        }
    }
    return %result;
}

sub ini_push ($) {
    my ($section) = @_;
    croak "unable to push nonexistent section [$section]"
        unless exists $inihash->{$section};
    push @inistack, $section;
}

sub ini_pop () {
    croak "can't pop empty INI section stack"
        unless @inistack;
    pop @inistack;
}

1;
